@@ -161,7 +161,7 @@ Worker::Worker(struct ev_loop *loop, SSL_CTX *sv_ssl_ctx, SSL_CTX *cl_ssl_ctx,
 #endif // ENABLE_HTTP3 && HAVE_LIBBPF
       randgen_(util::make_mt19937()),
       worker_stat_{},
-      dns_tracker_(loop),
+      dns_tracker_(loop, get_config()->conn.downstream->family),
 #ifdef ENABLE_HTTP3
       quic_upstream_addrs_{get_config()->conn.quic_listener.addrs},
 #endif // ENABLE_HTTP3
@@ -360,6 +360,10 @@ void Worker::replace_downstream_config(
         addr.seq = seq++;
       }
 
+      util::shuffle(std::begin(shared_addr->addrs),
+                    std::end(shared_addr->addrs), randgen_,
+                    [](auto i, auto j) { std::swap((*i).seq, (*j).seq); });
+
       if (shared_addr->affinity.type == SessionAffinity::NONE) {
         std::map<StringRef, WeightGroup *> wgs;
         size_t num_wgs = 0;
@@ -915,8 +919,9 @@ int Worker::create_quic_server_socket(UpstreamAddr &faddr) {
 
       rv = bpf_object__load(obj);
       if (rv != 0) {
+        auto error = errno;
         LOG(FATAL) << "Failed to load bpf object file: "
-                   << xsi_strerror(-rv, errbuf.data(), errbuf.size());
+                   << xsi_strerror(error, errbuf.data(), errbuf.size());
         close(fd);
         return -1;
       }
@@ -934,9 +939,9 @@ int Worker::create_quic_server_socket(UpstreamAddr &faddr) {
 
       ref.obj = obj;
 
-      auto reuseport_array =
+      ref.reuseport_array =
           bpf_object__find_map_by_name(obj, "reuseport_array");
-      if (!reuseport_array) {
+      if (!ref.reuseport_array) {
         auto error = errno;
         LOG(FATAL) << "Failed to get reuseport_array: "
                    << xsi_strerror(error, errbuf.data(), errbuf.size());
@@ -944,10 +949,8 @@ int Worker::create_quic_server_socket(UpstreamAddr &faddr) {
         return -1;
       }
 
-      ref.reuseport_array = bpf_map__fd(reuseport_array);
-
-      auto cid_prefix_map = bpf_object__find_map_by_name(obj, "cid_prefix_map");
-      if (!cid_prefix_map) {
+      ref.cid_prefix_map = bpf_object__find_map_by_name(obj, "cid_prefix_map");
+      if (!ref.cid_prefix_map) {
         auto error = errno;
         LOG(FATAL) << "Failed to get cid_prefix_map: "
                    << xsi_strerror(error, errbuf.data(), errbuf.size());
@@ -955,8 +958,6 @@ int Worker::create_quic_server_socket(UpstreamAddr &faddr) {
         return -1;
       }
 
-      ref.cid_prefix_map = bpf_map__fd(cid_prefix_map);
-
       auto sk_info = bpf_object__find_map_by_name(obj, "sk_info");
       if (!sk_info) {
         auto error = errno;
@@ -969,11 +970,12 @@ int Worker::create_quic_server_socket(UpstreamAddr &faddr) {
       constexpr uint32_t zero = 0;
       uint64_t num_socks = config->num_worker;
 
-      rv =
-          bpf_map_update_elem(bpf_map__fd(sk_info), &zero, &num_socks, BPF_ANY);
+      rv = bpf_map__update_elem(sk_info, &zero, sizeof(zero), &num_socks,
+                                sizeof(num_socks), BPF_ANY);
       if (rv != 0) {
+        auto error = errno;
         LOG(FATAL) << "Failed to update sk_info: "
-                   << xsi_strerror(-rv, errbuf.data(), errbuf.size());
+                   << xsi_strerror(error, errbuf.data(), errbuf.size());
         close(fd);
         return -1;
       }
@@ -984,20 +986,25 @@ int Worker::create_quic_server_socket(UpstreamAddr &faddr) {
       auto &qkms = conn_handler_->get_quic_keying_materials();
       auto &qkm = qkms->keying_materials.front();
 
-      rv = bpf_map_update_elem(bpf_map__fd(sk_info), &key_high_idx,
-                               qkm.cid_encryption_key.data(), BPF_ANY);
+      rv = bpf_map__update_elem(sk_info, &key_high_idx, sizeof(key_high_idx),
+                                qkm.cid_encryption_key.data(),
+                                qkm.cid_encryption_key.size() / 2, BPF_ANY);
       if (rv != 0) {
+        auto error = errno;
         LOG(FATAL) << "Failed to update key_high_idx sk_info: "
-                   << xsi_strerror(-rv, errbuf.data(), errbuf.size());
+                   << xsi_strerror(error, errbuf.data(), errbuf.size());
         close(fd);
         return -1;
       }
 
-      rv = bpf_map_update_elem(bpf_map__fd(sk_info), &key_low_idx,
-                               qkm.cid_encryption_key.data() + 8, BPF_ANY);
+      rv = bpf_map__update_elem(sk_info, &key_low_idx, sizeof(key_low_idx),
+                                qkm.cid_encryption_key.data() +
+                                    qkm.cid_encryption_key.size() / 2,
+                                qkm.cid_encryption_key.size() / 2, BPF_ANY);
       if (rv != 0) {
+        auto error = errno;
         LOG(FATAL) << "Failed to update key_low_idx sk_info: "
-                   << xsi_strerror(-rv, errbuf.data(), errbuf.size());
+                   << xsi_strerror(error, errbuf.data(), errbuf.size());
         close(fd);
         return -1;
       }
@@ -1017,20 +1024,23 @@ int Worker::create_quic_server_socket(UpstreamAddr &faddr) {
       const auto &ref = quic_bpf_refs[faddr.index];
       auto sk_index = compute_sk_index();
 
-      rv =
-          bpf_map_update_elem(ref.reuseport_array, &sk_index, &fd, BPF_NOEXIST);
+      rv = bpf_map__update_elem(ref.reuseport_array, &sk_index,
+                                sizeof(sk_index), &fd, sizeof(fd), BPF_NOEXIST);
       if (rv != 0) {
+        auto error = errno;
         LOG(FATAL) << "Failed to update reuseport_array: "
-                   << xsi_strerror(-rv, errbuf.data(), errbuf.size());
+                   << xsi_strerror(error, errbuf.data(), errbuf.size());
         close(fd);
         return -1;
       }
 
-      rv = bpf_map_update_elem(ref.cid_prefix_map, cid_prefix_.data(),
-                               &sk_index, BPF_NOEXIST);
+      rv = bpf_map__update_elem(ref.cid_prefix_map, cid_prefix_.data(),
+                                cid_prefix_.size(), &sk_index, sizeof(sk_index),
+                                BPF_NOEXIST);
       if (rv != 0) {
+        auto error = errno;
         LOG(FATAL) << "Failed to update cid_prefix_map: "
-                   << xsi_strerror(-rv, errbuf.data(), errbuf.size());
+                   << xsi_strerror(error, errbuf.data(), errbuf.size());
         close(fd);
         return -1;
       }
